package org.claros.commons.mail.protocols;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.ConnectException;
import java.util.ArrayList;
import java.util.Properties;

import javax.mail.FetchProfile;
import javax.mail.Flags;
import javax.mail.Folder;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.NoSuchProviderException;
import javax.mail.Session;
import javax.mail.Store;
import javax.mail.internet.MimeMessage;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.claros.commons.models.AuthProfile;
import org.claros.commons.exception.SystemException;
import org.claros.commons.mail.exception.ConnectionException;
import org.claros.commons.mail.exception.MailboxActionException;
import org.claros.commons.mail.exception.ServerDownException;
import org.claros.commons.mail.models.ConnectionMetaHandler;
import org.claros.commons.mail.models.ConnectionProfile;
import org.claros.commons.mail.models.EmailHeader;
import org.claros.commons.mail.utility.Constants;
import org.claros.commons.mail.utility.Utility;
import org.claros.commons.utility.Formatter;

/**
 * @author Umut Gokbayrak
 */
public class ImapProtocolImpl implements Protocol {
	private static Log log = LogFactory.getLog(ImapProtocolImpl.class);
	private String folder;
	private ConnectionProfile profile;
	private AuthProfile auth;
	private ConnectionMetaHandler handler;

	/**
	 * 
	 * @param profile
	 * @param auth
	 * @param handler
	 */
	ImapProtocolImpl(ConnectionProfile profile, AuthProfile auth, ConnectionMetaHandler handler, String folder) {
		this.profile = profile;
		this.auth = auth;
		this.handler = handler;
		this.folder = folder;
		if (this.folder == null || this.folder.trim().equals("")) {
			this.folder = Constants.FOLDER_INBOX;
		}
		
		if (this.folder.toLowerCase().equals(Constants.FOLDER_INBOX.toLowerCase())) {
			this.folder = this.folder.toUpperCase();
		}
		
	}

	/**
	 * 
	 * @return
	 * @throws Exception
	 */
	public Folder getFolder() throws Exception {
		Folder myFold = null;
		if (folder == null) {
			folder = Constants.FOLDER_INBOX;
		}

		if (folder != null && handler != null) {
			Store store = handler.getStore();
			if (store == null || !store.isConnected()) {
				log.info("Connection is closed. Restoring it...");
				connect(Constants.CONNECTION_READ_WRITE);
				log.info("Connection re-established");
			}
			myFold = handler.getStore().getFolder(folder);
			if (!myFold.isOpen() && myFold.exists()) {
				log.info("Folder :" + folder + " is closed. Opening again.");
				myFold.open(Constants.CONNECTION_READ_WRITE);
				log.info("Folder is open again.");
			}
		}
		return myFold;
	}

	/* (non-Javadoc)
	 * @see org.claros.commons.mail.protocols.Protocol#connect(int)
	 */
	public ConnectionMetaHandler connect(int connectType) throws SystemException, ConnectionException, ServerDownException {
		Folder fold = null;
		try {
			if (handler == null || handler.getStore() == null || !handler.getStore().isConnected()) {
				Properties props = new Properties();
				Session session = Session.getDefaultInstance(props);
				log.debug("session instance initiated");
				handler = new ConnectionMetaHandler();
				handler.setStore(session.getStore(profile.getProtocol()));
				log.debug("session store set. protocol is: " + profile.getProtocol());
				handler.getStore().connect(profile.getFetchServer(), profile.getIFetchPort(), auth.getUsername(), auth.getPassword());
				if (handler.getStore().isConnected()) {
					log.debug("Store has been connected... Successful");
				} else {
					log.warn("Connection unsuccessfull...!!");
				}
			}
			fold = handler.getStore().getFolder(Constants.FOLDER_INBOX);
			handler.setMbox(fold);
			log.debug("Got mailbox folder. Folder is: " + fold.getFullName());
			handler.setTotalMessagesCount(fold.getMessageCount());
			log.debug("Message Count:" + handler.getTotalMessagesCount());

			log.debug("OK. Mailbox got and checked for open. Now closing.");
		} catch (NoSuchProviderException e) {
			log.fatal(profile.getProtocol() + " provider could not be found.");
			throw new SystemException(e);
		} catch (MessagingException e) {
			Exception ne = e.getNextException();
			if (ne != null) {
				if (ne instanceof ConnectException || ne instanceof IOException) {
					throw new ServerDownException("Server is unreachable.");
				}
			}
			log.error("Connection could not be established.");
//			throw new ConnectionException(e);
		} catch (Exception e) {
			log.error("An unknown exception while connect.", e);
		} finally {
			closeFolder(fold);
		}
		return handler;
	}

	/* (non-Javadoc)
	 * @see org.claros.commons.mail.protocols.Protocol#deleteMessages(int[])
	 */
	public ConnectionMetaHandler deleteMessages(int[] messageIds) throws MailboxActionException, SystemException, ConnectionException {
		Folder fold = null;
		try {
			fold = getFolder();
			if (messageIds != null && messageIds.length > 0) {
				for (int i=0;i<messageIds.length;i++) {
					Message msg = fold.getMessage(messageIds[i]);
					msg.setFlag(Flags.Flag.DELETED, true);
				}
				fold.expunge();
			}
		} catch (MessagingException e) {
			log.error("Could not delete message ids: " + messageIds, e);
			throw new MailboxActionException(e);
		} catch (IndexOutOfBoundsException e) {
			log.error("Maybe you are double clicking the delete button", e);
		} catch (Exception e) {
			log.error("Could not delete message ids: " + messageIds, e);
			throw new MailboxActionException(e);
		} finally {
			closeFolder(fold);
		}
		return handler;
	}


	/**
	 * Fetches all e-mail headers from the server, with appropriate
	 * fields already set.
	 * @param handler
	 * @return ArrayList of MessageHeaders
	 * @throws ConnectionException
	 */
	public ArrayList fetchAllHeaders() throws SystemException, ConnectionException {
		ArrayList headers = new ArrayList();
		Folder fold = null;
		try {
			fold = getFolder();
			EmailHeader header = null;

			Message[] msgs = fold.getMessages();
			FetchProfile fp = new FetchProfile();
			fp.add(FetchProfile.Item.ENVELOPE);
			fp.add(FetchProfile.Item.FLAGS);
			fp.add(FetchProfile.Item.CONTENT_INFO);
			fp.add("Size");
			fp.add("Date");
			fold.fetch(msgs, fp);

			Message msg = null;
			for (int i = 0; i < msgs.length; i++) {
				try {
					header = new EmailHeader();
					msg = msgs[i];

					header.setMultipart((msg.isMimeType("multipart/*")) ? true : false);
					header.setMessageId(i + 1);
					header.setFrom(msg.getFrom());
					header.setTo(msg.getRecipients(Message.RecipientType.TO));
					header.setCc(msg.getRecipients(Message.RecipientType.CC));
					header.setBcc(msg.getRecipients(Message.RecipientType.BCC));
					header.setDate(msg.getSentDate());
					header.setReplyTo(msg.getReplyTo());
					header.setSize(msg.getSize());
					header.setSubject(msg.getSubject());
                    
					// now set the human readables.
					header.setDateShown(Formatter.formatDate(header.getDate(), "dd.MM.yyyy HH:mm"));
					header.setFromShown(Utility.addressArrToString(header.getFrom()));
					header.setToShown(Utility.addressArrToString(header.getTo()));
					header.setCcShown(Utility.addressArrToString(header.getCc()));
					header.setSizeShown(Utility.sizeToHumanReadable(header.getSize()));
                    
					if (profile.getProtocol().equals(Constants.IMAP)) {
						Flags.Flag flags[] = msg.getFlags().getSystemFlags();
						if (flags != null) {
							Flags.Flag flag = null;
							for (int m=0; m < flags.length; m++) {
								flag = flags[m];
								if (flag.equals(Flags.Flag.SEEN)) {
									header.setUnread(new Boolean(false));
								}
							}
						}
					}
					if (header.getUnread() == null) {
						header.setUnread(new Boolean(true));
					}
                    
					// it is time to add it to the arraylist
					headers.add(header);
				} catch (MessagingException e1) {
					log.error("Could not parse headers of e-mail. Message might be defuncted or illegal formatted.", e1);
				}
			}
		} catch (MessagingException e) {
			log.error("Could not fetch message headers. Is mbox connection still alive???", e);
//			throw new ConnectionException(e);
		} catch (Exception e) {
			log.error("Could not fetch message headers. Is mbox connection still alive???", e);
//			throw new ConnectionException(e);
		} finally {
			closeFolder(fold);
		}
		return headers;
	}


	/**
	 * Fetches and returns message headers as message objects.
	 * @return
	 * @throws SystemException
	 * @throws ConnectionException
	 */
	public ArrayList fetchAllHeadersAsMessages() throws SystemException, ConnectionException {
		ArrayList headers = null;
		Folder fold = null;
		try {
			headers = new ArrayList();
			fold = getFolder();
			Message[] msgs = fold.getMessages();
			FetchProfile fp = new FetchProfile();
			fp.add(FetchProfile.Item.ENVELOPE);
			fp.add(FetchProfile.Item.FLAGS);
			fp.add(FetchProfile.Item.CONTENT_INFO);
			fp.add("Size");
			fp.add("Date");
			fold.fetch(msgs, fp);

			Message msg = null;
			for (int i = 0; i < msgs.length; i++) {
				msg = msgs[i];
				headers.add(msg);
			}
		} catch (MessagingException e) {
			log.error("Could not fetch message headers. Is mbox connection still alive???", e);
			throw new ConnectionException(e);
		} catch (Exception e) {
			log.error("Could not fetch message headers. Is mbox connection still alive???", e);
			throw new ConnectionException(e);
		} finally {
			closeFolder(fold);
		}
		return headers;
	}

	public Message getMessage(int messageId) throws MailboxActionException, SystemException, ConnectionException, Exception {
		Message msg = null;
		Folder fold = null;
		try {
			try {
				fold = getFolder();
				msg = fold.getMessage(messageId);
			} catch (MessagingException e) {
				log.error("Could not fetch message body from remote server.", e);
				throw new MailboxActionException(e);
			}
		} catch (Exception e) {
			throw e;
		} finally {
			closeFolder(fold);
		}
		return msg;
	}

	/**
	 * Disconnects the previously opened data connection if
	 * the connection is still alive.
	 * @param handler
	 */
	public void disconnect() {
		if (handler != null && handler.getStore() != null) {
			try {
				handler.getStore().close();
			} catch (Exception e) {}
		}
	}

	public void markAsRead(Long msgId) throws Exception {
		Folder f = getFolder();
		try {
			Message msg = f.getMessage(msgId.intValue());
			msg.setFlag(Flags.Flag.SEEN, true);
		} catch (MessagingException e) {
			log.warn("Marking as Read not worked.", e);
		} finally {
			closeFolder(f);
		}
	}
  
	public void appendEmail(byte[] buff) throws Exception {
		Properties props = new Properties();
		Session session = Session.getDefaultInstance(props);
		ByteArrayInputStream bis = new ByteArrayInputStream(buff);
		MimeMessage msg = new MimeMessage(session, bis);
		msg.setFlag(Flags.Flag.SEEN, true);

		ProtocolFactory factory = new ProtocolFactory(profile, auth, handler);
		ImapProtocolImpl imap = (ImapProtocolImpl)factory.getImap(folder);
		Folder f = imap.getFolder();

		try {
			f.appendMessages(new Message[] {msg});
		} catch (MessagingException e) {
			log.warn("appenging msg to folder : " + folder + " failed.", e);
		} finally {
			closeFolder(f);
		}
	}

	public void moveEmail(Long msgId, String destFolder) throws Exception {
		ProtocolFactory factory = new ProtocolFactory(profile, auth, handler);
		ImapProtocolImpl fromProtocol = (ImapProtocolImpl)factory.getImap(folder);
		ImapProtocolImpl destProtocol = (ImapProtocolImpl)factory.getImap(destFolder);
		Folder from = fromProtocol.getFolder();
		Folder dest = null;

		try {
			Message msg = fromProtocol.getMessage(msgId.intValue());
			if (msg != null) {
				// because of the buggy imap servers lost the connection after getMessage
				// we need to check if the folder is open or not. 
				// (Do not use uw-imapd, it sucks!!!)
				from = fromProtocol.getFolder();
				dest = destProtocol.getFolder();
				from.copyMessages(new Message[] {msg}, dest);
				deleteMessages(new int[] {msg.getMessageNumber()});
			}
		} catch (IndexOutOfBoundsException e) {
			log.debug("Index kat. Moving message to folder : " + destFolder + " failed.", e);
		} catch (Exception e) {
			log.warn("Moving message to folder : " + destFolder + " failed.", e);
		} finally {
			closeFolder(from);
			closeFolder(dest);
		}
	}

	/**
	 * @return
	 */
	public Folder[] listFolders() throws Exception {
		ProtocolFactory factory = new ProtocolFactory(profile, auth, handler);
		ImapProtocolImpl protocol = (ImapProtocolImpl)factory.getImap(folder);
		Folder f = protocol.getFolder();

		Folder parent = null;
		Folder[] folders = null;
		
		try {
			parent = f.getParent();
			folders = parent.listSubscribed("*");
			/*
			if (folders == null) {
				Folder tmp = null;
				String n = null;
				for (int i=0; i<folders.length; i++) {
					tmp = folders[i];
					n = tmp.getName();
					
					if ((n == null || n.startsWith(".") || 
						tmp.getType() != Folder.HOLDS_MESSAGES) && !n.toUpperCase().equals(Constants.FOLDER_INBOX)) {
						folders[i] = null;
					}
				}
			} else {
				folders = new Folder[1];
				folders[0] = protocol.getFolder();
			}
			*/
		} catch (MessagingException e) {
			log.warn("Cannot get folder list.");
		} finally {
			closeFolder(f);
			closeFolder(parent);
		}
		return folders;
	}

	public static void closeFolder(Folder f) {
		if (f != null) {
			try {
				if (f.isOpen()) {
					f.close(true);
					log.info("Folder: " + f.getName() + " was open and now closed.");
				} else {
					log.info("Folder: " + f.getName() + " was already closed.");
				}
			} catch (MessagingException e) {
				log.info("Error while closing folder: " + f.getName(), e);
			}
		}
	}

	/**
	 * 
	 */
	public void createFolder() throws Exception {
		Folder f = getFolder();
		try {
			if (!f.exists()) {
				f.create(Folder.HOLDS_MESSAGES);
				f.setSubscribed(true);
			}
		} catch (MessagingException e) {
			log.warn("Could not create folder: " + f.getName());
		} finally {
			closeFolder(f);
		}
	}

	/**
	 * @return
	 */
	public int getUnreadMessageCount() throws Exception {
		Folder f = getFolder();

		if (f.exists()) {
			return f.getUnreadMessageCount();
		}
		closeFolder(f);
		return 0;
	}

	/**
	 * @return
	 */
	public int getTotalMessageCount() throws Exception {
		Folder f = getFolder();

		if (f.exists()) {
			return f.getMessageCount();
		}
		closeFolder(f);
		return 0;
	}

	/**
	 * 
	 */
	public void emptyFolder() throws Exception {
		Folder f = getFolder();

		try {
			Message msgs[] = f.getMessages();
			FetchProfile fp = new FetchProfile();
			fp.add(FetchProfile.Item.ENVELOPE);
			f.fetch(msgs, fp);
			
			int ids[] = new int[msgs.length];
			for (int i=0; i<msgs.length; i++) {
				ids[i] = msgs[i].getMessageNumber();
			}
			if (ids.length > 0) {
				deleteMessages(ids);
			}
		} catch (Exception e) {
			log.warn("Could not delete all messages in folder: " + folder);
		} finally {
			closeFolder(f);
		}
	}

	/**
	 * 
	 */
	public void deleteFolder() throws Exception {
		Folder f = getFolder();
		f.setSubscribed(false);
		closeFolder(f);
		f.delete(true);
	}

	/**
	 * @param newName
	 */
	public void renameFolder(String newName) throws Exception {
		Folder fOld = getFolder();
		Folder fNew = handler.getStore().getFolder(newName);
		closeFolder(fOld);
		fOld.renameTo(fNew);
	}

}
