package org.claros.commons.mail.protocols;

import java.util.ArrayList;
import java.util.Properties;

import javax.mail.AuthenticationFailedException;
import javax.mail.FetchProfile;
import javax.mail.Flags;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.NoSuchProviderException;
import javax.mail.Session;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.claros.commons.models.AuthProfile;
import org.claros.commons.exception.SystemException;
import org.claros.commons.mail.exception.ConnectionException;
import org.claros.commons.mail.exception.MailboxActionException;
import org.claros.commons.mail.exception.ServerDownException;
import org.claros.commons.mail.models.ConnectionMetaHandler;
import org.claros.commons.mail.models.ConnectionProfile;
import org.claros.commons.mail.models.EmailHeader;
import org.claros.commons.mail.utility.Constants;
import org.claros.commons.mail.utility.Utility;
import org.claros.commons.utility.Formatter;

/**
 * @author Umut Gokbayrak
 */
public class Pop3ProtocolImpl implements Protocol {
	private static Log log = LogFactory.getLog(Pop3ProtocolImpl.class);
	private ConnectionProfile profile;
	private AuthProfile auth;
	private ConnectionMetaHandler handler;

	/**
	 * 
	 * @param profile
	 * @param auth
	 * @param handler
	 */
	Pop3ProtocolImpl(ConnectionProfile profile, AuthProfile auth, ConnectionMetaHandler handler) {
		this.profile = profile;
		this.auth = auth;
		this.handler = handler;
	}

	/* (non-Javadoc)
	 * @see org.claros.commons.mail.protocols.FetchProtocol#connect(int)
	 */
	public ConnectionMetaHandler connect(int connectType) throws SystemException, ConnectionException, ServerDownException {
		try {
			if (handler == null || !handler.getStore().isConnected()) {
				Properties props = new Properties();
				Session session = Session.getDefaultInstance(props);
				log.debug("session instance initiated");
				handler = new ConnectionMetaHandler();
				handler.setStore(session.getStore(profile.getProtocol()));
				log.debug("session store set");
				handler.getStore().connect(profile.getFetchServer(), profile.getIFetchPort(), auth.getUsername(), auth.getPassword());
				log.debug("Store has been connected... Successful");
				handler.setMbox(handler.getStore().getDefaultFolder());
				handler.setMbox(handler.getMbox().getFolder(Constants.FOLDER_INBOX));
				log.debug("Got mailbox");
				handler.getMbox().open(connectType);
				log.debug("Mailbox open");
				handler.setTotalMessagesCount(handler.getMbox().getMessageCount());
				log.debug("Message Count:" + handler.getTotalMessagesCount());
			}
		} catch (AuthenticationFailedException e) {
			log.debug("I guess, mailbox was busy(read write lock). But user needn't know about this. Do nothing.", e);
		} catch (NoSuchProviderException e) {
			log.fatal(profile.getProtocol() + " provider could not be found.");
			throw new SystemException(e);
		} catch (MessagingException e) {
			disconnect();
			log.error("Connection could not be established.");
			throw new ConnectionException(e);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return handler;
	}

	/* (non-Javadoc)
	 * @see org.claros.commons.mail.protocols.FetchProtocol#deleteMessages(int[])
	 */
	public ConnectionMetaHandler deleteMessages(int[] messageIds) throws MailboxActionException, SystemException, ConnectionException {
		try {
			if (messageIds != null && messageIds.length > 0) {
				for (int i=0;i<messageIds.length;i++) {
					Message msg = handler.getMbox().getMessage(messageIds[i]);
					msg.setFlag(Flags.Flag.DELETED, true);
				}
			}
			disconnect();
			return handler;
		} catch (MessagingException e) {
			disconnect();
			log.error("Could not delete message ids: " + messageIds, e);
			throw new MailboxActionException(e);
		}
	}

	/**
	 * Fetches all e-mail headers from the server, with appropriate
	 * fields already set.
	 * @param handler
	 * @return ArrayList of MessageHeaders
	 * @throws ConnectionException
	 */
	public ArrayList fetchAllHeaders() throws SystemException, ConnectionException {
		ArrayList headers = null;
		try {
			headers = new ArrayList();
			EmailHeader header = null;
            
			Message[] msgs = handler.getMbox().getMessages();
			FetchProfile fp = new FetchProfile();
			fp.add(FetchProfile.Item.ENVELOPE);
			fp.add(FetchProfile.Item.FLAGS);
			fp.add(FetchProfile.Item.CONTENT_INFO);
			fp.add("Size");
			fp.add("Date");
			handler.getMbox().fetch(msgs, fp);

			Message msg = null;
			for (int i = 0; i < msgs.length; i++) {
				try {
					header = new EmailHeader();
					msg = msgs[i];

					header.setMultipart((msg.isMimeType("multipart/*")) ? true : false);
					header.setMessageId(i + 1);
					header.setFrom(msg.getFrom());
					header.setTo(msg.getRecipients(Message.RecipientType.TO));
					header.setCc(msg.getRecipients(Message.RecipientType.CC));
					header.setBcc(msg.getRecipients(Message.RecipientType.BCC));
					header.setDate(msg.getSentDate());
					header.setReplyTo(msg.getReplyTo());
					header.setSize(msg.getSize());
					header.setSubject(msg.getSubject());
                    
					// now set the human readables.
					header.setDateShown(Formatter.formatDate(header.getDate(), "dd.MM.yyyy HH:mm"));
					header.setFromShown(Utility.addressArrToString(header.getFrom()));
					header.setToShown(Utility.addressArrToString(header.getTo()));
					header.setCcShown(Utility.addressArrToString(header.getCc()));
					header.setSizeShown(Utility.sizeToHumanReadable(header.getSize()));
                    
					// it is time to add it to the arraylist
					headers.add(header);
				} catch (MessagingException e1) {
					log.debug("Could not parse headers of e-mail. Message might be defuncted or illegal formatted.", e1);
				}
			}
//			  disconnect();
		} catch (MessagingException e) {
			disconnect();
			log.error("Could not fetch message headers. Is mbox connection still alive???", e);
			throw new ConnectionException(e);
		}
		return headers;
	}

	/**
	 * Fetches and returns message headers as message objects.
	 * @return
	 * @throws SystemException
	 * @throws ConnectionException
	 */
	public ArrayList fetchAllHeadersAsMessages() throws SystemException, ConnectionException {
		ArrayList headers = null;
		try {
			headers = new ArrayList();
            
			Message[] msgs = handler.getMbox().getMessages();
			FetchProfile fp = new FetchProfile();
			fp.add(FetchProfile.Item.ENVELOPE);
			fp.add(FetchProfile.Item.FLAGS);
			fp.add(FetchProfile.Item.CONTENT_INFO);
			fp.add("Size");
			fp.add("Date");
			handler.getMbox().fetch(msgs, fp);

			Message msg = null;
			for (int i = 0; i < msgs.length; i++) {
				msg = msgs[i];
				headers.add(msg);
			}
		} catch (MessagingException e) {
			disconnect();
			log.error("Could not fetch message headers. Is mbox connection still alive???", e);
			throw new ConnectionException(e);
		}
		return headers;
	}

	public Message getMessage(int messageId) throws MailboxActionException, SystemException, ConnectionException, Exception {
		Message msg = null;
		try {
			try {
				msg = handler.getMbox().getMessage(messageId);
			} catch (MessagingException e) {
				log.error("Could not fetch message body from remote server.", e);
				throw new MailboxActionException(e);
			}
		} catch (Exception e) {
			throw e;
		}
		return msg;
	}

	/**
	 * Disconnects the previously opened data connection if
	 * the connection is still alive.
	 * @param handler
	 */
	public void disconnect() {
		if (handler != null) {
			try {
				if (handler.getMbox() != null) {
					handler.getMbox().close(true);
				}
			} catch (Exception e) {}
			try {
				if (handler.getStore() != null) {
					handler.getStore().close();
				}
			} catch (Exception e) {}
		}
	}
}
